@@ -5,6 +5,7 @@ import shutil
 import string
 import random
 import argparse
+import threading
 from getpass import getuser
 from traceback import format_exception
 from collections import OrderedDict
@@ -48,6 +49,9 @@ SYSTEMD_RUN_UNIT_NAME = 'cfy_cluster_manager_{}'
 BASE_CFY_DIR = '/etc/cloudify/'
 INITIAL_INSTALL_DIR = join(BASE_CFY_DIR, '.installed')
 
+UPGRADE_RPM_PATH = 'http://repository.cloudifysource.org/cloudify/5.1.1/ga-' \
+                   'release/cloudify-manager-install-5.1.1-ga.el7.x86_64.rpm'
+
 
 class CfyNode(VM):
     def __init__(self,
@@ -77,6 +81,13 @@ class CfyNode(VM):
             BASE_CFY_DIR, '{}_config.yaml'.format(node_name))
         self.unit_name = SYSTEMD_RUN_UNIT_NAME.format(self.type)
 
+    def get_version(self):
+        # You need to verify cloudify-manager-install is installed
+        cfy_version_res = self.run_command(
+            'rpm --queryformat "%{VERSION}" -q cloudify-manager-install',
+            hide_stdout=True)
+        return cfy_version_res.stdout
+
 
 def _exception_handler(type_, value, traceback):
     error = type_.__name__
@@ -342,7 +353,7 @@ def _verify_cloudify_installed_successfully(instance):
             'Service {} status is unknown'.format(instance.unit_name))
 
 
-def _install_instances(instances_dict, verbose, override):
+def _install_instances(instances_dict, verbose):
     for i, instance_type in enumerate(instances_dict):
         logger.info('installing %s instances', instance_type)
         for instance in instances_dict[instance_type]:
@@ -493,11 +504,13 @@ def _log_managers_connection_strings(manager_nodes):
                 'following connection strings:\n%s', managers_str)
 
 
-def _print_successful_installation_message(start_time, msg='installed'):
+def _print_success_message(start_time, msg='installed'):
     running_time = time.time() - start_time
     m, s = divmod(running_time, 60)
     logger.info('Cloudify cluster was successfully {0} in '
                 '{1} minutes and {2} seconds'.format(msg, int(m), int(s)))
+    logger.info('Please run `cfy cluster status` to verify the cluster status '
+                'is healthy. It might take up to a minute for it to stabilize')
 
 
 def _install_cloudify_locally(rpm_path):
@@ -814,11 +827,7 @@ def _remove_cloudify_installation(instance, verbose):
     instance.run_command(
         'rm -f {0}'.format(instance.config_path), use_sudo=True)
 
-    if not _are_any_services_installed(instance):
-        instance.run_command(
-            'yum remove -y cloudify-manager-install', use_sudo=True)
-
-    if instance.type == 'manager':
+    if '5.1.0' in instance.get_version() and instance.type == 'manager':
         certs_paths_list = [
             'cloudify_external_cert.pem', 'cloudify_external_key.pem',
             'cloudify_internal_ca_cert.pem', 'cloudify_external_ca_cert.pem',
@@ -833,11 +842,24 @@ def _remove_cloudify_installation(instance, verbose):
                 instance.run_command(
                     'mv {0} {1}'.format(full_path, new_path), use_sudo=True)
 
+    if not _are_any_services_installed(instance):
+        instance.run_command(
+            'yum remove -y cloudify-manager-install', use_sudo=True)
+
     instance.run_command('rm -rf {}'.format(CLUSTER_INSTALL_DIR))
 
     instance.installed = False
 
 
+def _get_reversed_instances_dict(instances_dict):
+    reversed_instances_dict = OrderedDict(reversed(
+        list(instances_dict.items())))
+    for instance_type, instances in reversed_instances_dict.items():
+        reversed_instances_dict[instance_type] = list(reversed(instances))
+
+    return reversed_instances_dict
+
+
 def _handle_installed_instances(instances_dict, override, verbose):
     """Checking which instances were installed in the previous installation.
 
@@ -860,7 +882,8 @@ def _handle_installed_instances(instances_dict, override, verbose):
     removed its installation. This is the useful if the user wants to continue
     the installation from where it previously stopped.
     """
-    for instance_type, instances_list in instances_dict.items():
+    reversed_instances_dict = _get_reversed_instances_dict(instances_dict)
+    for instance_type, instances_list in reversed_instances_dict.items():
         for instance in instances_list:
             logger.info('Checking if %s was installed', instance.name)
             if instance.file_exists(instance.config_path):
@@ -875,7 +898,9 @@ def _handle_installed_instances(instances_dict, override, verbose):
                 else:
                     logger.info('Previous Cloudify installation of %s failed',
                                 instance.name)
-                    _create_installation_files(instance, verbose)
+                    if '5.1.0' in instance.get_version():
+                        _create_installation_files(instance, verbose)
+
                     logger.info('Removing failed Cloudify installation '
                                 'from %s', instance.name)
                     _remove_cloudify_installation(instance, verbose)
@@ -883,10 +908,6 @@ def _handle_installed_instances(instances_dict, override, verbose):
                         continue
 
                     return
-            else:
-                if override:
-                    continue
-                return
 
 
 def install(config_path, override, only_validate, verbose):
@@ -924,14 +945,14 @@ def install(config_path, override, only_validate, verbose):
             credentials = _handle_credentials(config.get('credentials'))
         _prepare_config_files(instances_dict, credentials, config)
 
-    _install_instances(instances_dict, verbose, override)
+    _install_instances(instances_dict, verbose)
     _log_managers_connection_strings(instances_dict['manager'])
     if credentials:
         logger.warning('The credentials file was saved to %s. '
                        'The credentials are written there in plain text. '
                        'Please remove it after reviewing it.',
                        CREDENTIALS_FILE_PATH)
-    _print_successful_installation_message(start_time)
+    _print_success_message(start_time)
 
 
 def remove(config_path, verbose):
@@ -947,15 +968,98 @@ def remove(config_path, verbose):
                       if using_three_nodes_cluster else
                       _generate_general_cluster_dict(config))
 
-    previous_installation = _previous_installation(instances_dict)
-    if previous_installation:
+    if _previous_installation(instances_dict):
         _handle_installed_instances(instances_dict, True, verbose)
-        _print_successful_installation_message(start_time, 'removed')
+        _print_success_message(start_time, 'removed')
     else:
         logger.info('No previous installation of a Cloudify cluster was '
                     'detected. Nothing to remove.')
 
 
+def _upgrade_cluster(instances_dict, verbose, upgrade_rpm_path,
+                     using_three_nodes_cluster):
+    instances_list = instances_dict['manager']
+    if not using_three_nodes_cluster:
+        instances_list += (instances_dict['rabbitmq'] +
+                           instances_dict['postgresql'])
+
+    _install_upgrade_rpm_on_nodes(instances_list, upgrade_rpm_path)
+
+    for instance_type, instances_list in instances_dict.items():
+        for instance in instances_list:
+            logger.info('Upgrading %s', instance.name)
+            logger.info('Running upgrade command on %s', instance.name)
+            instance.run_command(
+                'cfy_manager upgrade -c {config} {verbose}'.format(
+                    config=instance.config_path,
+                    verbose='-v' if verbose else '')
+            )
+
+
+def _install_upgrade_rpm_on_nodes(instances_list, upgrade_rpm_path):
+    threads = []
+    for i, instance in enumerate(instances_list, start=1):
+        new_thread = threading.Thread(target=_thread_rpm_upgrade,
+                                      args=(instance, upgrade_rpm_path,))
+        threads.append(new_thread)
+        new_thread.start()
+
+    for i, thread in enumerate(threads):
+        thread.join()
+        logger.info('Finished installing upgrade RPM on %s',
+                    instances_list[i].name)
+
+
+def _thread_rpm_upgrade(instance, rpm_path):
+    logger.info('Installing upgrade RPM on %s', instance.name)
+    instance.run_command('yum install -y {} --disablerepo=*'.format(rpm_path),
+                         use_sudo=True, hide_stdout=True)
+
+
+def _should_upgrade(instances_dict, using_three_nodes_cluster):
+    logger.info('Verifying v5.1.0 is installed on all instances')
+    for instance_type, instances_list in instances_dict.items():
+        for instance in instances_list:
+            logger.debug('Verifying instance %s', instance.private_ip)
+            if not _rpm_was_installed(instance):
+                logger.debug('cloudify-manager-install is not installed on %s',
+                             instance.private_ip)
+                return False
+            instance_version = instance.get_version()
+            if '5.1.0' not in instance_version:
+                logger.debug(
+                    'cloudify-manager-install version on instance %s is not '
+                    '5.1.0. It is %s', instance.private_ip, instance_version)
+                return False
+        if using_three_nodes_cluster:  # One iteration goes through all nodes
+            return True
+
+    return True
+
+
+def upgrade(config_path, verbose, upgrade_rpm_path):
+    if not yum_is_present():
+        raise ClusterInstallError('Yum is not present.')
+
+    start_time = time.time()
+    logger.info('Upgrading Cloudify cluster')
+    config_path = config_path or CLUSTER_INSTALL_CONFIG_PATH
+    config = get_dict_from_yaml(config_path)
+    using_three_nodes_cluster = (len(config.get('existing_vms')) == 3)
+    instances_dict = (_generate_three_nodes_cluster_dict(config)
+                      if using_three_nodes_cluster else
+                      _generate_general_cluster_dict(config))
+
+    if _should_upgrade(instances_dict, using_three_nodes_cluster):
+        _upgrade_cluster(instances_dict, verbose, upgrade_rpm_path,
+                         using_three_nodes_cluster)
+        _print_success_message(start_time, 'upgraded')
+    else:
+        logger.info("Couldn't upgrade cluster. Please verify "
+                    "cloudify-manager-install v5.1.0 is installed on all "
+                    "cluster nodes")
+
+
 def add_verbose_arg(parser):
     parser.add_argument(
         '-v', '--verbose',
@@ -1046,6 +1150,22 @@ def main():
     add_config_arg(remove_args)
     add_verbose_arg(remove_args)
 
+    upgrade_args = subparsers.add_parser(
+        'upgrade',
+        help='Upgrade a v5.1.0 Cloudify cluster to v5.1.1')
+
+    add_config_arg(upgrade_args)
+    upgrade_args.add_argument(
+        '--upgrade-rpm',
+        action='store',
+        default=UPGRADE_RPM_PATH,
+        help='Path to a v5.1.1 cloudify-manager-install RPM. '
+             'This can be either a local or remote path. '
+             'Default: {0}'.format(UPGRADE_RPM_PATH)
+    )
+
+    add_verbose_arg(upgrade_args)
+
     args = parser.parse_args()
 
     if hasattr(args, 'verbose'):
@@ -1061,6 +1181,9 @@ def main():
     elif args.action == 'remove':
         remove(args.config_path, args.verbose)
 
+    elif args.action == 'upgrade':
+        upgrade(args.config_path, args.verbose, args.upgrade_rpm)
+
     else:
         raise RuntimeError('Invalid action specified in parser.')
 
