@@ -1,7 +1,6 @@
 import os
 import re
 import shlex
-import socket
 import subprocess
 from os.path import dirname, exists, expanduser, isdir, isfile, join
 from socket import error as socket_error
@@ -85,25 +84,32 @@ class VM(object):
                  private_ip,
                  public_ip,
                  key_file_path,
-                 username):
+                 username,
+                 password=None):
         self.username = username
         self.private_ip = private_ip
         self.public_ip = public_ip or private_ip
-        self.key_file_path = expanduser(key_file_path)
+        self.key_file_path = (expanduser(key_file_path) if key_file_path
+                              else None)
+        self.password = password if password else None
 
     def _get_connection(self):
+        connect_kwargs = ({'key_filename': [self.key_file_path]} if
+                          self.key_file_path else {'password': self.password})
         connection = Connection(
             host=self.private_ip, user=self.username, port=22,
-            connect_kwargs={'key_filename': [self.key_file_path]})
+            connect_kwargs=connect_kwargs)
         self.test_connection(connection)
 
         return connection
 
     def test_connection(self, connection=None):
         """ Connection is lazy, so **we** need to check it can be opened."""
+        connect_kwargs = ({'key_filename': [self.key_file_path]} if
+                          self.key_file_path else {'password': self.password})
         connection = connection or Connection(
             host=self.private_ip, user=self.username, port=22,
-            connect_kwargs={'key_filename': [self.key_file_path]})
+            connect_kwargs=connect_kwargs)
         try:
             connection.open()
         except (socket_error, AuthenticationException) as exc:
@@ -142,7 +148,7 @@ class VM(object):
                          local_path, remote_path, self.private_ip)
             connection.put(expanduser(local_path), remote_path)
 
-    def put_dir(self, local_dir_path, remote_dir_path, override=False):
+    def put_dir(self, local_dir_path, remote_dir_path):
         """Copy a local directory to a remote host.
 
         This function wraps the recursive function _put_dir(). This way
@@ -151,23 +157,19 @@ class VM(object):
         :param local_dir_path: An existing local directory path.
         :param remote_dir_path: A directory path on the remote host. If the
                                 path doesn't exist, it will be created.
-        :param override: If True and the remote directory path exists, then
-                         it will be deleted.
-                         If False and the remote directory path exists, then
-                         the files from the local directory will be added to
-                         the remote one.
-                         If the remote directory path doesn't exist, it
-                         doesn't have any effect.
         """
         if not isdir(local_dir_path):
             raise ClusterInstallError(
                 '{} is not a directory'.format(local_dir_path))
 
-        logger.debug('Copying %s to %s on host %s',
-                     local_dir_path, remote_dir_path, self.private_ip)
-        if override:
-            self.run_command('rm -rf {}'.format(remote_dir_path))
-        self._put_dir(self._get_connection(), local_dir_path, remote_dir_path)
+        if self.file_exists(remote_dir_path):
+            logger.debug('The files already exist on instance %s',
+                         self.private_ip)
+        else:
+            logger.debug('Copying %s to %s on host %s',
+                         local_dir_path, remote_dir_path, self.private_ip)
+            self._put_dir(self._get_connection(), local_dir_path,
+                          remote_dir_path)
 
     def _put_dir(self, connection, local_dir_path, remote_dir_path):
         connection.run('mkdir -p {}'. format(remote_dir_path), warn=True,
@@ -212,10 +214,6 @@ def yum_is_present():
         return False
 
 
-def current_host_ip():
-    return socket.gethostbyname(socket.getfqdn())
-
-
 def openssl_command(file_name, file_format='x509', extra_flags_list=None):
     command = ['openssl', file_format, '-in', file_name, '-noout']
     if extra_flags_list:
